<?php
/**
 * @file nodeapi functions for date_repeat_sequence
 */
/**
 * Updates a nodes date_repeats sequences
 */
function _date_repeat_sequence_node_update(&$node) {
  $sequences = $node->date_repeat_sequence;
  foreach($sequences as $field => $values) {
    if (empty($node->date_repeat_sequences[$field])) {
      if (empty($values)) {
        //if there are no values then don't do anything
        break;
      }
      _date_repeat_sequence_node_insert($node, $field);
    }
    else {
      $update_op = empty($node->date_repeat_sequence_edit[$field]) ? 'all' : $node->date_repeat_sequence_edit[$field];
      $function = "_date_repeat_sequence_node_update_" . $update_op;
      $args = array(
        $node,
        $field,
      );
      call_user_func_array($function, $args);
    }
  }
}

/**
 * Inserts a nodes date_repeats sequences
 *
 * @param $node
 *   The node object.
 *
 * @param $field_name
 *   The field name to insert a sequence for. If none is specified then all sequence fields will be inserted.
 */
function _date_repeat_sequence_node_insert(&$node, $field_name = NULL) {
  $sequences = $node->date_repeat_sequence;
  if (!empty($field_name) && isset($sequences[$field_name])) {
    $seq = array(
     'pnid' => $node->nid,
     'field' => $field_name,
    );
    drupal_write_record('date_repeat_sequence', $seq);
    $node->date_repeat_sequences[$seq['field']] = $seq;
    $record = array('sqid' => $seq['sqid'], 'nid' => $node->nid);
    drupal_write_record('date_repeat_sequence_node', $record);
    _date_repeat_sequence_generate_nodes($seq, $node);
    return;
  }
  foreach($sequences as $field => $values) {
    $seq = array(
     'pnid' => $node->nid,
     'field' => $field,
    );
    drupal_write_record('date_repeat_sequence', $seq);
    $node->date_repeat_sequences[$seq['field']] = $seq;
    $record = array('sqid' => $seq['sqid'], 'nid' => $node->nid);
    drupal_write_record('date_repeat_sequence_node', $record);
    _date_repeat_sequence_generate_nodes($seq, $node);
  }
}

/**
 * Generates nodes that are a part of a sequence.
 */
function _date_repeat_sequence_generate_nodes($seq, $node) {
  $sequences = $node->date_repeat_sequence;
  $field = $seq['field'];
  foreach ($sequences[$field] as $occurance) {
    $new_node = drupal_clone($node);
    $new_node->$field = array($occurance);
    unset($new_node->date_repeat_sequence);
    unset($new_node->nid);
    node_save($new_node);
    $record = array('sqid' => $seq['sqid'], 'nid' => $new_node->nid);
    drupal_write_record('date_repeat_sequence_node', $record);
    unset($new_node);
  }
}

/**
 * Saves an existing repeat sequence for all nodes in the sequence
 *
 * @param $node
 *   The Node object
 *
 * @param $field
 *   The sequence Field name
 */
function _date_repeat_sequence_node_update_all($node, $field) {
  $sqid = $node->date_repeat_sequences[$field]['sqid'];
  $pnid = $node->date_repeat_sequences[$field]['pnid'];
  $nids = date_repeat_sequence_load_nids($sqid);
  $values = $node->date_repeat_sequence[$field];
  $rev =!empty($node->revision);
  if ($pnid != $node->nid) {
    //In the case where the user changed the time or end of an event in a non-parent node then we need to regenerate the
    //Parent nodes dates and and times and put those into the values.
    $parent_node = node_load($pnid);
    $diff = 0;
    $val_date = array_shift($values);
    $rrule = $val_date['rrule'];
    $val_start = date_make_date($val_date['value'], $val_date['timezone']);

    if (!empty($val_date['value2'])) {
      $val_end = date_make_date($val_date['value2'], $val_date['timezone']);
      $diff = date_difference($val_start, $val_end);
    }

    $val_start = date_array($val_start);
    $parent_value = array_shift($parent_node->$field);
    $parent_start = date_array(date_make_date($parent_value['value']));
    $parent_start['second'] = $val_start['second'];
    $parent_start['minute'] = $val_start['second'];
    $parent_start['hour'] = $val_start['hour'];
    $parent_end = $parent_start;

    if($diff) {
      $parent_end['second'] = $parent_end['second'] + $diff;
    }

    $parent_value['value'] = date_format_date(date_make_date($parent_start, $parent_value['timezone'], DATE_ARRAY), 'custom', DATE_FORMAT_DATETIME);
    $parent_value['value2'] = date_format_date(date_make_date($parent_end, $parent_value['timezone'], DATE_ARRAY), 'custom', DATE_FORMAT_DATETIME);
    $parent_value['rrule'] = $rrule;
    $fields = content_fields();
    $date_field = $fields[$field];
    module_load_include('inc', 'date', 'date_repeat');
    module_load_include('inc', 'date_api', 'date_api_ical');
    $rrule = date_ical_parse_rrule('',$rrule);
    $values = date_repeat_build_dates(NULL,$rrule, $date_field, $parent_value);
    $parent_node = drupal_clone($node);
    $parent_node->nid = $pnid;
    $parent_node->$field = array(array_shift($values));
    $parent_node->date_repeat_sequence[$field] = $values;
    node_save($parent_node);
    return;
  }
  //We need to remove the pnid from the group as it is already loded with it's date
  unset($nids[$pnid]);
  foreach ($nids as $nid) {
    $sequence_node = node_load($nid);
    $nodes[$nid] = $sequence_node;
    $value = array_shift($sequence_node->$field);
    $field_values[$nid]['value'] = $value['value'];
    unset($field_values[$nid]['rrule']);
  }
  ksort($nodes);
  foreach ($values as $value) {
    //create a search value that we can unset the rrule on.
    $search_value['value'] = $value['value'];
    unset($search_value['rrule']);
    $nid = array_search($search_value, $field_values);
    if($nid) {
      //There is already a node with this date.
      //This prevents us from constantly creating and deleting nodes.
      //We will alwyas try to use all available nodes before creating or deleting nodes.
      //So now we must copy all values from the old node into this node.
      $new_node = drupal_clone($node);
      $new_node->nid = $nid;
      $new_node->$field = array($value);
      unset($new_node->date_repeat_sequence[$field]);
      if ($rev) {
        unset($new_node->vid);
      }
      else {
        $vid = db_result(db_query('SELECT vid FROM {node} WHERE nid = %d', $nid));
        $new_node->vid = $vid;
      }
      $ready_nodes[$nid] = drupal_clone($new_node);
      unset($field_values[$nid]);
      unset($nodes[$nid]);
      unset($new_node);
    }
    else {      
      //there is not an existing node with this value so we will save our new one for now
      //and assign it one of the available nids later.
      $new_node = drupal_clone($node);
      unset($new_node->nid);
      unset($new_node->vid);
      unset($new_node->date_repeat_sequence[$field]);
      $new_node->$field = array($value);
      $new_nodes[] = drupal_clone($new_node);
      unset($new_node);
    }
  }

  //At this point we should have one or two arrays of nodes. Either $new_nodes, $ready_nodes or both.
  //We also have $field_values that is loaded with the left over $nids we can grab.
  if (count($field_values) > 0 && is_array($new_nodes)) {
    //if there are nids left to use and nodes that need them then we will use them
    foreach ($field_values as $nid => $old) {
      $ready_nodes[$nid] = array_shift($new_nodes);
      $ready_nodes[$nid]->nid = $nid;
      unset($field_values[$nid]);
      unset($nodes[$nid]);
    }
  }
  while (count($new_nodes) > 0) {
    //if after iterating through the remaining nids available we still have $new repeat nodes to add
    //then we iterate through them and add them to the $ready_node array without a $nid
    $ready_nodes[] = array_shift($new_nodes);
  }
  //at this point we now have all updated and new nodes saved in $ready_nodes
  //So now we will save all the $ready_nodes and then delete any remaining nodes in the $nodes array.
  foreach ($ready_nodes as $nid => $save_node) {
    $new = NULL;
    $new = empty($save_node->nid);
    node_save($save_node);
    if ($new) {
      //if the $save_node->nid  was preexisting then we simpley updated and don't need to add a new record to the sequence table
      //If the $save_node->nid was empty then it was a new node and needs to be saved to the sequence table
      $record = array('sqid' => $sqid, 'nid' => $save_node->nid);
      drupal_write_record('date_repeat_sequence_node', $record);
    }
  }
  //Next if there are any nodes left in the $nodes array they need to be deleted.
  if (count($nodes) > 0) {
    foreach ($nodes as $nid => $delete_node) {
      node_delete($nid);
    }
  }
}

/**
 * Removes a node from the sequence it is in and resets the rrule for all other nodes in that sequence.
 *
 * @param $node
 *   The Node object
 *
 * @param $field
 *   The sequence Field name
 */
function _date_repeat_sequence_node_update_only($node, $field) {
  $sqid = $node->date_repeat_sequences[$field]['sqid'];
  $nids = date_repeat_sequence_load_nids($sqid);
  if(count($nids) < 2) {
    //do nothing as this is the only node in this sequence already.
    return;
  }
  if ($node->nid == $node->date_repeat_sequences[$field]['pnid']) {
    //this is the parent node so we do not need to update the rrule just remove the parent node and shift the
    //Sequence parent to the next node in the sequence
    date_repeat_sequence_shift_parent($node->date_repeat_sequences[$field]);
  }
  //We can delete this node from the sequence
  db_query('DELETE FROM {date_repeat_sequence_node} WHERE sqid = %d AND nid = %d', $sqid, $node->nid);

  //also remove the rrule from the db
  $content_db = content_database_info(content_fields($field));
  db_query('UPDATE {%s} SET %s = "" WHERE vid = %d', $content_db['table'], $content_db['columns']['rrule']['column'], $node->vid);
}

/**
 * Saves an existing repeat sequence for all nodes in the future of the sequence
 *
 * @param $node
 *   The Node object
 *
 * @param $field
 *   The sequence Field name
 */
function _date_repeat_sequence_node_update_future($node, $field) {
  $sqid = $node->date_repeat_sequences[$field]['sqid'];
  module_load_include('inc', 'date_api', 'date_api_sql');
  $value = array_shift($node->$field);
  $date_sql = new date_sql_handler;
  $content_db = content_database_info(content_fields($field));
  $table = $content_db['table'];
  $column = array_shift($content_db['columns']);
  $sql = 'SELECT n.nid FROM {date_repeat_sequence_node} dn
    JOIN {node} n ON n.nid = dn.nid
    JOIN {' . $table . '} c ON c.vid = n.vid
    ORDER BY c.' . $column['column'] . ' ASC
    WHERE dn.sqid = %d';
 $sql .= ' AND ' . $date_sql->sql_where_date('FIELD', 'c.' . $column['column'], '>', $value['value']);

 $results = db_query($sql, $sqid);
 while ($nid = db_result($results)) {
   $nids[$nid] = $nid;
 }
 //Create our new sequence.
 $seq = array(
   'pnid' => $node->nid,
   'field' => $field,
 );
 drupal_write_record('date_repeat_sequence', $seq);
 //Update the nodes in this sequence
 db_query('UPDATE {date_repeat_sequence_node} SET sqid = %d WHERE sqid = %d AND nid IN (%s)', $seq['sqid'], $sqid, implode(',',$nids));

 $node->date_repeat_sequences[$field] = $seq;
 //now lets run it through the update_all function for this new sequence
 _date_repeat_sequence_node_update_all($node, $field);
}

/**
 * Deletes a nodes date_repeats sequences
 */
function _date_repeat_sequence_node_delete(&$node) {
  $sequences = $node->date_repeat_sequences;
  if (!is_array($sequences)) {
    return;
  }
  foreach($sequences as $field => $values) {
    $update_op = $node->date_repeat_sequence_edit[$field];
    $sqid = $node->date_repeat_sequences[$field]['sqid'];
    $seq = $node->date_repeat_sequences[$field];
    if ($seq['pnid'] == $node->nid && $update_op == 'future') {
      //if this is the parent node and the command was to delete this and all future nodes
      //then we need to kill the whole sequence
      $update_op = 'all';
    }
    switch ($update_op) {
      case 'all':
          $nids = date_repeat_sequence_load_nids($sqid);
          foreach ($nids as $nid) {
            if ($nid != $node->nid) {
              //so we don't get stuck in an infinite loop
              node_delete($nid);
            }
          }
          db_query('DELETE FROM {date_repeat_sequence} WHERE sqid = %d', $sqid);
          db_query('DELETE FROM {date_repeat_sequence_node} WHERE sqid = %d', $sqid);
        break;
      case 'only':
      default:
        db_query('DELETE FROM {date_repeat_sequence_node} WHERE sqid = %d AND nid = %d', $sqid, $node->nid);
        break;
      case 'future':
        module_load_include('inc', 'date_api', 'date_api_sql');
        $value = array_shift($node->$field);
        $date_sql = new date_sql_handler;
        $content_db = content_database_info(content_fields($field));
        $table = $content_db['table'];
        $column = array_shift($content_db['columns']);
        $sql = 'SELECT n.nid FROM {date_repeat_sequence_node} dn
          JOIN {node} n ON n.nid = dn.nid
          JOIN {' . $table . '} c ON c.vid = n.vid
          ORDER BY c.' . $column['column'] . ' ASC
          WHERE dn.sqid = %d';
        $sql .= ' AND ' . $date_sql->sql_where_date('FIELD', 'c.' . $column['column'], '>=', $value['value']);

        $results = db_query($sql, $sqid);
        while ($nid = db_result($results)) {
          $nids[$nid] = $nid;
        }
         db_query('DELETE FROM {date_repeat_sequence_node} WHERE sqid = %d AND nid IN ($s)', $sqid, implode(',', $nids));
        break;
    }
  }
}

/**
 * Shifts the node sequence off of the current parent and on to the next date in that sequence
 *
 * @param $seq
 *   The sequence array. Keys are sqid, pnid, & field.
 */
function date_repeat_sequence_shift_parent($seq) {
  $content_db = content_database_info(content_fields($seq['field']));
  $table = $content_db['table'];
  $column = array_shift($content_db['columns']);
  $sql = 'SELECT n.nid FROM {date_repeat_sequence_node} dn
    JOIN {node} n ON n.nid = dn.nid
    JOIN {' . $table . '} c ON c.vid = n.vid
    WHERE dn.sqid = %d AND dn.nid != %d
    ORDER BY c.' . $column['column'] . ' ASC';
  $nid = db_result(db_query($sql, $seq['sqid'], $seq['pnid']));
  $seq['pnid'] = $nid;
  drupal_write_record('date_repeat_sequence', $seq, 'sqid');
}
